/* Copyright, 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Search all trees using gentreeg for trees of isomorphic halves with >
   2^(n/2) minimal domsets.

   -------------
   h=17
   2 = 5-middle
   3 = 5-attachment
   11 = 4-middle
   12 = 4-attachment

   found 47530 v=2 u=2  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,19,2,19,20,19,22,19,24,19,26,18,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}c@Gsi\IjQjsjLEv[lvj]B
   :ag?GKIFCapg{aRIdrHkwu[Mrr@ctMIDQp]]OGcaYB
   count 132400 > target 131072
   --
   found 47530 v=2 u=3  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,19,3,19,20,19,22,19,24,19,26,18,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cAGsi\IjQjsjLEv[lvj]B
   :afah[qZSBqHKEDBaPWsA[YLfIXOiUhdr@czMFcAHGz
   count 131609 > target 131072
   --
   found 47530 v=2 u=11  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,19,11,19,20,19,22,19,24,19,26,18,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cIGsi\IjQjsjLEv[lvj]B
   :afah[qZSBPxCe@@`OwcU[WKeiXkw\hdAhWnMFcAHG~
   count 133480 > target 131072
   --
   found 47530 v=2 u=12  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,19,12,19,20,19,22,19,24,19,26,18,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cJGsi\IjQjsjLEv[lvj]B
   :afah[qZSCQWKIFCapg{A[YLfIXOi\hdax_rMFcAHGz
   count 131637 > target 131072
   --
   found 47530 v=3 u=11  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,19,20,11,20,19,22,19,24,19,26,18,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cPdCi\IjQjsjLEv[lvj]B
   :afAxcucGcqgKIFCapg|s@MFcALcsZMFYhWmWMSQPKg\
   count 133081 > target 131072
   --
   found 47530 v=11 u=11  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,28,18,19,20,19,22,19,24,19,26,11,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cPhSyTUdVhUXPEv[lvj]B
   :ag?GKIFCapg{aRIdrHkwgTJDxx?aQHrbXozNKERQB
   count 134432 > target 131072
   --
   found 47530 v=11 u=12  n=34 0,1,2,3,2,5,2,7,2,9,1,11,12,11,14,11,16,28,18,19,20,19,22,19,24,19,26,12,28,29,28,31,28,33
   :a_OWSEH@bOXCATJdRXS}cPhSyTUdVhUXXEv[lvj]B
   :afBHlCPHdQwKIFCapg|s@MFcBlcsZMFYhWmWMSAHGeS
   count 133141 > target 131072
   --
   h=17 end, tree count 48629
   max minimal_domsets=134432 reps 1 target_reps 7
   elapsed time 16 = 0:16

   --
   gentreeg  num_vertices=18 target_count=262144
   h=18 end, tree count 123867
   max minimal_domsets=262144 reps 3807 target_reps 0
   elapsed time 90 = 1:30
   --
   h=19 end, tree count 317955
   max minimal_domsets=551392 reps 4 target_reps 123
   elapsed time 303 = 5:03

*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <nauty/nauty.h>
#include <nauty/nausparse.h>
#include <nauty/naututil.h>
#include <nauty/gtools.h>
#include <nauty/gutils.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int my_geng_main(int argc, char**argv);
int my_gentreeg_main(int argc, char**argv);

const int option_verbose = 1;

unsigned long long  trees_total_domnum[64];

#define MAX(x,y)  ((x) > (y) ? (x) : (y))
#define MIN(x,y)  ((x) < (y) ? (x) : (y))


unsigned long  target_count;
unsigned long  good_num = 0;
unsigned long long  tree_count = 0;
unsigned long  max_value = 0;
int            max_reps = 0;
int            target_reps = 0;
#define  TARGET_STRS_SIZE  10000
char *target_strs[TARGET_STRS_SIZE];
SG_DECL(sg);
SG_DECL(canon_sg);

void
my_gentreeg_outproc (FILE *outfile, int *half_vpar, int h)
{
  DEBUG(printf("my_gentreeg_outproc() h=%d\n", h));
  assert(half_vpar[1] == 0);  /* root=1 from gentreeg */
  tree_count++;

  int half_rerooted[MAXN+1];
  assert (h <= MAXN);
  memcpy(half_rerooted, half_vpar, (h+1)*sizeof(int));
  nautyextra_validate_vpar(half_rerooted,h);

  int vpar[MAXN+1];
  assert (h <= MAXN);
  memcpy(vpar, half_vpar, (h+1)*sizeof(int));
  nautyextra_validate_vpar(vpar,h);

  int n = 2*h;
  assert (n <= MAXN);

  int v;
  for (v=1; v<=h; v++) {
    DEBUG(printf(" v=%d\n", v));
    nautyextra_reroot_vpar(half_rerooted,h, v);
    assert (half_rerooted[v] == 0);
    assert(nautyextra_num_roots_vpar(half_rerooted,h) == 1);

    int u;
    for (u=v; u<=h; u++) {
      DEBUG(printf("  u=%d\n", u));
      assert(nautyextra_num_roots_vpar(vpar,h) == 1);
      assert(vpar[1] == 0);  /* root=1 from gentreeg */

      nautyextra_join_vpar(vpar,h, u, half_rerooted,h);
      nautyextra_validate_vpar(vpar,n);
      assert(nautyextra_num_roots_vpar(vpar,n) == 1);

      unsigned long  count_minimal_domsets
        = nautyextra_minimal_domsets_count_vpar(vpar,n);

      if (count_minimal_domsets > max_value) {
        max_value = count_minimal_domsets;
        max_reps = 0;
        if (0) {
          printf("n=%d new high max_value %lu target %lu  at tree_count %Lu\n",
                 n, max_value, target_count, tree_count);
        }
      }
      if (max_value == count_minimal_domsets) {
        max_reps++;
      }

      if (count_minimal_domsets > target_count) {
        if (target_reps >= TARGET_STRS_SIZE) {
          printf("oops TARGET_STRS_SIZE = %d exceeded\n", TARGET_STRS_SIZE);
          abort();
        }
        graph g[2*MAXN];
        int m = SETWORDSNEEDED(n);
        assert (m <= 2);
        nautyextra_vpar_to_graph(vpar,g,m,n, 0);

        graph canon_g[2*MAXN];
        nautyextra_canonise(g,m,n, canon_g);
        char *str = ntos6(canon_g,m,n);

        /* assert (nautyextra_is_undirected_sg(&sg)); */
        /* nautyextra_canonise_sg(&sg,&canon_sg); */
        /* nautyextra_vpar_to_sg(vpar,n, &sg, 0); */
        /* assert (nautyextra_is_undirected_sg(&sg)); */
        /* nautyextra_canonise_sg(&sg,&canon_sg); */
        /* assert (nautyextra_is_undirected_sg(&canon_sg)); */
        /* char *str = sgtos6(&canon_sg); */

        int i;
        for (i = 0; i < target_reps; i++)
          if (strcmp(str, target_strs[i]) == 0)
            goto SEEN_ALREADY;

        if (target_strs[target_reps]) free(target_strs[target_reps]);
        target_strs[target_reps++] = strdup(str);

        printf("found %Lu v=%d u=%d  ",tree_count,v,u);
        nautyextra_print_vpar(vpar,n);
        printf("%s", ntos6(g,m,n));
        printf("%s", ntos6(canon_g,m,n));
        /* printf("%s", sgtos6(&sg)); */
        /* printf("%s", sgtos6(&canon_sg)); */
        printf("count %lu > target %lu\n", count_minimal_domsets, target_count);
        printf("--\n");

      SEEN_ALREADY: ;
      }
    }
  }
}

int
main (void)
{
  setbuf(stdout,NULL);
  nautyextra_want_progress = 1;
  SG_INIT(sg);
  SG_INIT(canon_sg);

  int num_vertices;
  for (num_vertices = 1;
       num_vertices <= 17;
       num_vertices++) {
    tree_count = 0;
    target_count = 1 << num_vertices;
    max_value = 0;
    target_reps = 0;
    max_reps = 0;
    time_t start_time;
    time(&start_time);

    printf("gentreeg  num_vertices=%d target_count=%lu\n",
           num_vertices, target_count);
    char num_vertices_str[20];
    sprintf(num_vertices_str, "%d", num_vertices);

    int gentreeg_argc = 0;
    char *gentreeg_argv[10];
    gentreeg_argv[gentreeg_argc++] = "gentreeg"; /* progname */
    gentreeg_argv[gentreeg_argc++] = "-q";       /* quiet */
    gentreeg_argv[gentreeg_argc++] = num_vertices_str;
    gentreeg_argv[gentreeg_argc] = NULL;
    my_gentreeg_main(gentreeg_argc,gentreeg_argv);

    time_t end_time;
    time(&end_time);

    printf("h=%d end, tree count %Lu\n", num_vertices, tree_count);
    printf("  max minimal_domsets=%lu reps %d target_reps %d\n",
           max_value, max_reps, target_reps);

    time_t elapsed_time = end_time - start_time;
    printf("  elapsed time %ld = %ld:%02ld\n",
           elapsed_time, elapsed_time/60, elapsed_time%60);
    printf("\n");
  }

  return 0;
}
