/* Copyright 2018 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

/* Exercise nautyextra_make_random_vpar() with a fixed random number
   generator to see that all random sequences give different trees.
 */

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <nauty/gtools.h>

#include "nautyextra.h"
#include "nautyextra-wip.h"

#define WANT_DEBUG  0  /* set to 1 for debug prints */
#if WANT_DEBUG
#define DEBUG(expr)  expr
#else
#define DEBUG(expr)
#endif

int option_verbose = 0;

#define numberof(array)  (sizeof(array)/sizeof((array)[0]))

int count_total = 0;
int count_bad = 0;

const int max_bad = 1;

#define MAX_N  7
int random_array[MAX_N];
int random_array_size;
int random_upto;
int
nautyextra_test_random (int limit)
{
  if (random_upto >= random_array_size) {
    printf("random_upto %d past random_array_size %d\n",
           random_upto, random_array_size);
    abort();
  }
  int r = random_array[random_upto++];
  assert (0 <= r && r < limit);
  return r;
}

void
test_random_all (int n)
{
  assert (n <= MAX_N);
  /* if (option_verbose) { */
  /*   printf("tree vpar=%p  n=%d\n", vpar,n); */
  /*   nautyextra_print_vpar(vpar,n); */
  /* } */

  random_array_size = (n==0 ? 0 : n-1);
  int i;
  for (i = 0; i < numberof(random_array); i++) random_array[i] = -999;
  for (i = 0; i < n; i++) random_array[i] = 0;

  int forest;
  for (forest = 0; forest <= 1; forest++) {
    DEBUG(printf("n=%d forest=%d\n", n, forest));

    int limit = n + forest;  /* 0 to n-1 tree, 0 to n forest */
    int want_count = pow(limit, random_array_size);
    int count = 0;

    int max_key = 0;
    for (i = 1; i <= n; i++) {
      max_key *= n+1;
      max_key += n;
    }
    int seen_size = max_key + 1;
    DEBUG(printf("max_key %d seen_size %d\n", max_key, seen_size));
    int seen_m = SETWORDSNEEDED(seen_size);
    DYNALLSTAT(setword, seen, seen_sz);
    DYNALLOC1(setword,seen,seen_sz,seen_m,"test_random_all");
    EMPTYSET(seen,seen_m);

    for (;;) {
      DEBUG(printf("random[%d] ", random_array_size);
            for (i = 0; i < random_array_size; i++)
              printf(" %d", random_array[i]);
            printf("\n");
            );
      count++;

      int vpar[MAX_N+1];
      for (i = 0; i <= n; i++) vpar[i] = -999;

      random_upto = 0;
      nautyextra_make_random_vpar(vpar,n,forest);
      assert (random_upto == random_array_size);

      nautyextra_validate_vpar(vpar,n);

      /* distinct trees */
      int key = 0;
      for (i = 1; i <= n; i++) {
        key *= n+1;
        assert (0 <= vpar[i] && vpar[i] <= n);
        key += vpar[i];
      }
      DEBUG(printf("  key %d (max %d)\n", key, max_key));
      assert (0 <= key && key <= max_key);
      assert (! ISELEMENT(seen,key));
      ADDELEMENT(seen,key);

      /* increment array */
      int pos = 0;
      for (;;) {
        if (pos >= random_array_size) goto NEXT_ARRAY;
        if (++random_array[pos] < limit) break;
        random_array[pos++] = 0;
      }
    }
  NEXT_ARRAY: ;
    DEBUG(printf("  count %d want %d\n", count, want_count));
    assert (count == want_count);
    assert (count == setsize(seen,seen_m));
  }
}

int
main (int argc, char **argv)
{
  setbuf(stdout,NULL);

  int n;
  for (n = 0; n <= MAX_N; n++)
    test_random_all(n);

  return (count_bad ? 1 : 0);
}
