#!/usr/bin/perl -w

# Copyright 2017, 2018 Kevin Ryde

# This file is part of nautyextra.
#
# nautyextra is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 3, or (at your option) any later
# version.
#
# nautyextra is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with nautyextra.  If not, see <http://www.gnu.org/licenses/>.


use strict;
use FindBin;
use File::Spec;
use File::Slurp;
use Tie::IxHash;

# uncomment this to run the ### lines
# use Smart::Comments;

sub pos_linenum {
  my ($str, $pos) = @_;
  $pos //= pos($_[0]);
  $str = substr($str, 0, $pos);
  return 1 + scalar($str =~ tr/\n//);
}

my $bad = 0;

foreach my $elem
  (['nautyextra-wip.c','nautyextra-wip.h'],
   ['nautyextra.c',    'nautyextra.h']) {
  my ($c_filename, $h_filename) = @$elem;
  $c_filename = File::Spec->catfile($FindBin::Bin, File::Spec->updir,
                                    $c_filename);
  $h_filename = File::Spec->catfile($FindBin::Bin, File::Spec->updir,
                                    $h_filename);

  my $c_str = File::Slurp::read_file($c_filename);
  $c_str =~ s/  +/ /g;
  tie my %c_funcs, 'Tie::IxHash';
  while ($c_str =~ /(.*)\n(int )?(nautyextra_\w+)/mg) {
    my $type = $2 || $1;
    my $func = $3;
    ### c file ...
    ### $func
    ### $type
    next if $type =~ /^static\b/;
    $c_funcs{$func} = pos_linenum($c_str);
  }

  my $h_str = File::Slurp::read_file($h_filename);
  $h_str =~ s/  +/ /g;
  tie my %h_funcs, 'Tie::IxHash';
  while ($h_str =~ /^((const|char|int|unsigned|void|long|extern|NAUTYEXTRA_TYPE_ULL|graph|sparsegraph) |\*)+(nautyextra_\w+)/mg) {
    my $func = $3;
    ### h: $func
    $h_funcs{$func} = pos_linenum($h_str);;
  }
  delete $h_funcs{'nautyextra_test_random'}; # not in nautyextra-wip.c
  print "count $c_filename ",scalar(keys %c_funcs),
    "  $h_filename ",scalar(keys %h_funcs),"\n";

  foreach my $func (keys %c_funcs) {
    if (! $h_funcs{$func}) {
      print "$c_filename:$c_funcs{$func}: $func not in $h_filename\n";
      $bad++;
    }
  }
  foreach my $func (keys %h_funcs) {
    if (! $c_funcs{$func}) {
      print "$h_filename:$h_funcs{$func}: $func not in $c_filename\n";
      $bad++;
    }
  }
}

exit ($bad ? 1 : 0);
